import json

import lusid
import finbourne_access
import finbourne_identity

from finbourne_access import models as access_models
from finbourne_identity import models as identity_models


from finbourne_sdk_utils.cocoon.utilities import (
    checkargs,
)
import logging


@checkargs
def create_role(
    access_api_factory: finbourne_access.extensions.SyncApiClientFactory,
    identity_api_factory: finbourne_identity.extensions.SyncApiClientFactory,
    access_role_creation_request: access_models.RoleCreationRequest,
) -> None:
    """
    Creates a role through both the access and identity APIs

    Parameters
    ----------
    access_api_factory : finbourne_access.extensions.SyncApiClientFactory
    identity_api_factory : finbourne_identity.extensions.id.SyncApiClientFactory
        
    access_role_creation_request : access_models.RoleCreationRequest
        The role creation request to use

    Returns
    -------
    responses: None

    """
    
    access_roles_api = access_api_factory.build(finbourne_access.RolesApi)
    
    identity_roles_api = identity_api_factory.build(finbourne_identity.RolesApi)


    # Create the role using the access API.
    try:
        access_roles_api.create_role(access_role_creation_request)
        logging.info(
            f"Role with code {access_role_creation_request.code} has been created via the access API"
        )
    except finbourne_access.ApiException as e:
        detail = json.loads(e.body)
        if detail["code"] not in [612, 613, 615]:  # RoleWithCodeAlreadyExists
            raise e
        logging.info(
            f"Role with code {access_role_creation_request.code} has already been created via the access API"
        )

    # Create the same role using the identity API.
    identity_role_creation_request = identity_models.CreateRoleRequest(
        name=access_role_creation_request.code
    )
    try:
        identity_roles_api.create_role(identity_role_creation_request)
        logging.info(
            f"Role with code {access_role_creation_request.code} has been created via the identity API"
        )
    except finbourne_identity.ApiException as e:
        detail = json.loads(e.body)
        if detail["code"] != 157:  # RoleWithCodeAlreadyExists
            raise e
        logging.info(
            f"Role with code {access_role_creation_request.code} has already been created via the identity API"
        )



import lusid
import lusid.models as models
import logging

logger = logging.getLogger()


def create_transaction_type_configuration(api_factory, alias, movements):
    """
    This function creates a transaction type configuration if it doesn't already exist.

    Parameters
    ----------
    api_factory :  The LUSID api factory to use
    alias : lusid.models.TransactionConfigurationTypeAlias
        An aliases with type and group
    movements : list[lusid.models.TransactionConfigurationMovementDataRequest]
        The movements to use for
        transaction type

    Returns
    -------
    response : (lusid.models.createtransactiontyperesponse)
        The response from creating the transaction type
    """

    # Call LUSID to get your transaction type configuration
    response = api_factory.build(
        lusid.api.SystemConfigurationApi
    ).list_configuration_transaction_types()

    aliases_current = [
        (alias.type, alias.transaction_group)
        for transaction_grouping in response.transaction_configs
        for alias in transaction_grouping.aliases
    ]

    logger.info(
        f"The LUSID enviornment currently has {len(aliases_current)} transaction aliases"
    )

    if (alias.type, alias.transaction_group) in aliases_current:
        logging.warning(
            "The following alias already exists: "
            + f"Type of {alias.type} with source {alias.transaction_group}"
        )
        return response

    logger.info(f"Creating a new transaction aliases called: {alias.type}")

    response = api_factory.build(
        lusid.api.SystemConfigurationApi
    ).create_configuration_transaction_type(
        transaction_configuration_data_request=models.TransactionConfigurationDataRequest(
            aliases=[alias], movements=movements
        )
    )

    return response


def upsert_transaction_type_alias(api_factory, new_transaction_config):
    """
    This function overrides the current transaction types alias with a new configuration.
    The alias will be created if it does not already exist.

    Parameters
    ----------
    api_factory: factory
        The LUSID api factory to use.
    new_transaction_config: list[lusid.models.TransactionConfigurationDataRequest]
        A list of new transaction type configurations

    Returns
    -------
    response : (lusid.models.TransactionSetConfigurationData)
        The response from setting the transaction type
    """

    # Call LUSID to get a list of the current transaction types

    current_transaction_types = api_factory.build(
        lusid.api.SystemConfigurationApi
    ).list_configuration_transaction_types()

    transaction_configs_list = current_transaction_types.transaction_configs

    # Define function to delete current alias if it already exists

    def delete_current_alias(updated_alias):

        for txn_index, trans_type in enumerate(transaction_configs_list):

            for alias_index, alias in enumerate(trans_type.aliases):

                if (
                    alias.type == updated_alias.type
                    and alias.transaction_group == updated_alias.transaction_group
                ):
                    del trans_type.aliases[alias_index]

                    break

            else:
                continue

            # If there are no aliases assigned against the old movement, delete the transaction type configuration from LUSID
            # We don't want to leave unassigned movements in LUSID

            if len(trans_type.aliases) == 0:
                del transaction_configs_list[txn_index]

            break

    # Loop over list of new aliases and delete them if they already exist

    new_aliases_nested = [item.aliases for item in new_transaction_config]
    new_aliases = [item for sublist in new_aliases_nested for item in sublist]

    for item in new_aliases:
        delete_current_alias(item)

    # Then set the new aliases with new config

    transaction_configs_list = transaction_configs_list + new_transaction_config

    set_response = api_factory.build(
        lusid.api.SystemConfigurationApi
    ).set_configuration_transaction_types(
        transaction_set_configuration_data_request=models.TransactionSetConfigurationDataRequest(
            transaction_config_requests=transaction_configs_list,
            side_config_requests=current_transaction_types.side_definitions,
        )
    )

    return set_response



import lusid
import lusid.models as models
from lusid.api import InstrumentsApi, SearchApi
import numpy as np
import time
from finbourne_sdk_utils.cocoon.utilities import checkargs
import pandas as pd
import logging
import re
from finbourne_sdk_utils.cocoon.async_tools import run_in_executor
import asyncio
from typing import Callable


@checkargs
def prepare_key(identifier_lusid: str, full_key_format: bool) -> str:
    """
    This function prepares the key for the identifier based on whether the full key or just the code is required

    Parameters
    ----------
    identifier_lusid : str
        The LUSID identifier in either full key format or code only
    full_key_format : bool
        Whether they key needs to be in the full key format

    Returns
    -------
    str
        The LUSID identifier in the correct format
    """

    if full_key_format:
        return (
            identifier_lusid
            if re.findall("Instrument/\S+/\S+", identifier_lusid)
            else f"Instrument/default/{identifier_lusid}"
        )
    else:
        return (
            identifier_lusid.split("/")[2]
            if re.findall("Instrument/default/\S+", identifier_lusid)
            else identifier_lusid
        )


@checkargs
def create_identifiers(
    index,
    row: pd.Series,
    file_type: str,
    instrument_identifier_mapping: dict = None,
    unique_identifiers: list = None,
    full_key_format: bool = True,
    prepare_key: Callable = prepare_key,
) -> dict:
    """

    Parameters
    ----------
    index
        The index of the row in the DataFrame
    row : pd.Series
        The row of the DataFrame to create identifiers for
    file_type : str
        The file type to create identifiers for
    instrument_identifier_mapping : dict
        The instrument identifier mapping to use
    unique_identifiers : list
        The list of allowable unique instrument identifiers
    full_key_format : bool
        Whether the full key format i.e. 'Instrument/default/Figi' is required
    prepare_key : callable
        The function to use to prepare the key

    Returns
    -------
    identifiers : dict
        The identifiers to use on the request
    """

    # Populate the identifiers for this instrument
    identifiers = {
        # Handles specifying the entire property key e.g. Instrument/default/Figi or just the code e.g. Figi
        prepare_key(identifier_lusid, full_key_format): models.InstrumentIdValue(
            value=str(row[identifier_column])
        )
        if file_type == "instrument"
        else str(row[identifier_column])
        for identifier_lusid, identifier_column in instrument_identifier_mapping.items()
        if not pd.isna(  # Only use the identifier if it has a value
            row[identifier_column]  # type is instrument_identifiers: Dict[str, StrictStr]
        )
    }

    # If there are no identifiers raise an error
    if len(identifiers) == 0:
        raise ValueError(
            f"""The row at index {str(index)} has no value for every single one of the provided 
        identifiers. Please ensure that each row has at least one identifier and try again"""
        )

    # Check that at least one unique identifier exists if it is an instrument file (need to move this out of here)
    if file_type == "instrument":

        # Get the unique list of unique identifiers which have been populated
        unique_identifiers_populated = list(
            set(unique_identifiers).intersection(set(list(identifiers.keys())))
        )

        # If there are no unique identifiers raise an Exception as you need at least one to make a successful call
        if len(unique_identifiers_populated) == 0:
            raise ValueError(
                f"""The instrument at index {str(index)} has no value for at least one unique 
            identifier. Please ensure that each instrument has at least one unique identifier and try again. The
            allowed unique identifiers are {str(unique_identifiers)}"""
            )

    else:
        # If the transaction/holding is cash remove all other identifiers and just use this one
        if "Instrument/default/Currency" in list(identifiers.keys()):
            currency_value = identifiers["Instrument/default/Currency"]
            identifiers.clear()
                
            if currency_value == 'nan':
                currency_value = 'GBP'  # default to GBP if not supplied
        
            identifiers["Instrument/default/Currency"] = currency_value

    return identifiers


@checkargs
def resolve_instruments(
    api_factory: lusid.SyncApiClientFactory,
    data_frame: pd.DataFrame,
    identifier_mapping: dict,
):
    """
    This function attempts to resolve each row of the file to an instrument in LUSID

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        An instance of the Lusid Api Factory
    data_frame : pd.DataFrame
        The DataFrame containing the transactions or holdings to resolve to unique instruments
    identifier_mapping : dict
        The column mapping between allowable identifiers in LUSID and identifier columns in the dataframe

    Returns
    -------
    _data_frame : pd.DataFrame
        The input DataFrame with resolution columns added
    """

    if "Currency" not in list(
        identifier_mapping.keys()
    ) and "Instrument/default/Currency" not in list(identifier_mapping.keys()):
        raise KeyError(
            """There is no column specified in the identifier_mapping to identify whether or not an instrument is cash.
               Please specify add the key "Currency" or "Instrument/default/Currency" to your mapping. If no instruments
               are cash you can set the value to be None"""
        )

    if "Currency" in list(identifier_mapping.keys()):
        identifier_mapping["Instrument/default/Currency"] = identifier_mapping[
            "Currency"
        ]
        del identifier_mapping["Currency"]

    # Check that the values of the mapping exist in the DataFrame
    if not (set(identifier_mapping.values()) <= set(data_frame.columns)):
        raise KeyError(
            "there are values in identifier_mapping that are not columns in the dataframe"
        )

    # Get the allowable instrument identifiers from LUSID
    response = api_factory.build(InstrumentsApi).get_instrument_identifier_types()
    """
    # Collect the names and property keys for the identifiers and concatenate them
    allowable_identifier_names = [identifier.identifier_type for identifier in response.values]
    allowable_identifier_keys = [identifier.property_key for identifier in response.values]
    allowable_identifiers = allowable_identifier_names + allowable_identifier_keys

    # Check that the identifiers in the mapping are all allowed to be used in LUSID
    if not (set(identifier_mapping['identifier_mapping'].keys()) <= set(allowable_identifiers)):
        raise Exception(
            'there are LUSID identifiers in the identifier_mapping which are not configured in LUSID')
    """
    # Copy the data_frame to ensure the original isn't modified
    _data_frame = data_frame.copy(deep=True)

    # Set up the result Pandas Series to track resolution
    found_with = pd.Series(index=_data_frame.index, dtype=np.dtype(object))
    resolvable = pd.Series(index=_data_frame.index, dtype=np.dtype(bool))
    luid = pd.Series(index=_data_frame.index, dtype=np.dtype(object))
    comment = pd.Series(index=_data_frame.index, dtype=np.dtype(object))
    logging.info("Beginning instrument resolution process")
    # Iterate over each row in the DataFrame
    for index, row in _data_frame.iterrows():

        if index % 10 == 0:
            logging.info(f"Up to row {index}")
        # Initialise list to hold the identifiers used to resolve
        found_with_current = []
        # Initialise a value of False for the row's resolvability to an instrument in LUSID
        resolvable_current = False
        # Initialise the LUID value
        luid_current = None
        # Initialise the comment value
        comment_current = "No instruments found for the given identifiers"
        # Takes the currency resolution function and applies it
        currency = row[identifier_mapping["Instrument/default/Currency"]]

        if not pd.isna(currency):
            resolvable_current = True
            found_with_current.append(currency)
            luid_current = currency
            comment_current = "Resolved as cash with a currency"

        search_requests = [
            models.InstrumentSearchProperty(
                key=f"Instrument/default/{identifier_lusid}"
                if "Instrument/" not in identifier_lusid
                else identifier_lusid,
                value=str(row[identifier_dataframe]),
            )
            for identifier_lusid, identifier_dataframe in identifier_mapping.items()
            if not pd.isnull(row[identifier_dataframe])
        ]

        # Call LUSID to search for instruments
        attempts = 0

        if len(search_requests) > 0:
            while attempts < 3:
                try:
                    response = api_factory.build(SearchApi).instruments_search(
                        instrument_search_property=search_requests, mastered_only=True
                    )
                    break
                except lusid.exceptions.ApiException as error_message:
                    attempts += 1
                    comment_current = f"Failed to find instrument due to LUSID error during search due to status {error_message.status} with reason {error_message.reason}"
                    time.sleep(5)

            if attempts == 3:
                # Update the luid series
                luid.iloc[index] = luid_current
                # Update the found with series
                found_with.iloc[index] = found_with_current
                # Update the resolvable series
                resolvable.iloc[index] = resolvable_current
                # Update the comment series
                comment.iloc[index] = comment_current
                continue

            search_request_number = -1

            for result in response:

                search_request_number += 1
                # If there are matches
                if len(result.mastered_instruments) == 1:
                    # Add the identifier responsible for the successful search request to the list
                    found_with_current.append(
                        search_requests[search_request_number].key.split("/")[2]
                    )
                    comment_current = (
                        "Uniquely resolved to an instrument in the securities master"
                    )
                    resolvable_current = True
                    luid_current = (
                        result.mastered_instruments[0]
                        .identifiers["LusidInstrumentId"]
                        .value
                    )
                    break

                elif len(result.mastered_instruments) > 1:
                    comment_current = f'Multiple instruments found for the instrument using identifier {search_requests[search_request_number].key.split("/")[2]}'
                    resolvable_current = False
                    luid_current = np.nan

        # Update the luid series
        luid.iloc[index] = luid_current
        # Update the found with series
        found_with.iloc[index] = found_with_current
        # Update the resolvable series
        resolvable.iloc[index] = resolvable_current
        # Update the comment series
        comment.iloc[index] = comment_current

    # Add the series to the dataframe
    _data_frame["resolvable"] = resolvable
    _data_frame["foundWith"] = found_with
    _data_frame["LusidInstrumentId"] = luid
    _data_frame["comment"] = comment

    return _data_frame


@checkargs
def get_unique_identifiers(api_factory: lusid.SyncApiClientFactory):

    """
    Tests getting the unique instrument identifiers

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        The LUSID api factory to use

    Returns
    -------
    list[str]
        The property keys of the available identifiers
    """
    # Get the allowed instrument identifiers from LUSID
    identifiers = api_factory.build(
        lusid.api.InstrumentsApi
    ).get_instrument_identifier_types()

    # Return the identifiers that are configured to be unique
    return [
        identifier.identifier_type
        for identifier in identifiers.values
        if identifier.is_unique_identifier_type
    ]


async def enrich_instruments(
    api_factory: lusid.SyncApiClientFactory,
    data_frame: pd.DataFrame,
    instrument_identifier_mapping: dict,
    mapping_required: dict,
    constant_prefix: str = "$",
    **kwargs,
):
    search_requests_all = []

    for index, row in data_frame.iterrows():
        search_requests_instrument = [
            lusid.models.InstrumentSearchProperty(
                key=identifier_lusid
                if re.findall("Instrument/default/\S+", identifier_lusid)
                else f"Instrument/default/{identifier_lusid}",
                value=row[identifier_column],
            )
            for identifier_lusid, identifier_column in instrument_identifier_mapping.items()
            if not pd.isna(row[identifier_column])
        ]

        search_requests_all.append(search_requests_instrument)

    responses = await asyncio.gather(
        *[
            instrument_search(
                api_factory=api_factory, search_requests=search_requests, **kwargs
            )
            for search_requests in search_requests_all
        ],
        return_exceptions=False,
    )

    names = []

    for response in responses:
        name = np.nan
        for single_search in response:
            if isinstance(single_search, Exception):
                logging.warning(single_search)
                continue
            elif len(single_search[0].external_instruments) > 0:
                name = single_search[0].external_instruments[0].name
                break
        names.append(name)

    enriched_column_name = "LUSID.Name.Enriched"

    data_frame[enriched_column_name] = names

    # Missing mapping for name altogether
    if "name" not in list(mapping_required.keys()):
        mapping_required["name"] = enriched_column_name

    # A column for name already exists and needs to be enriched
    elif (
        isinstance(mapping_required["name"], str)
        and mapping_required["name"][0] != constant_prefix
    ):
        data_frame[mapping_required["name"]] = data_frame[
            mapping_required["name"]
        ].fillna(value=data_frame[enriched_column_name])

    elif isinstance(mapping_required["name"], dict) and "column" in list(
        mapping_required["name"].keys()
    ):
        data_frame[mapping_required["name"]["column"]] = data_frame[
            mapping_required["name"]["column"]
        ].fillna(value=data_frame[enriched_column_name])

    # Is a constant specified by the constant prefix
    elif (
        isinstance(mapping_required["name"], str)
        and mapping_required["name"][0] == constant_prefix
    ):
        mapping_required["name"] = {"default": mapping_required["name"][1:]}
        mapping_required["name"]["column"] = enriched_column_name

    # Is a constant specified by the default nested dictionary specification
    elif (
        isinstance(mapping_required["name"], dict)
        and "default" in list(mapping_required["name"].keys())
        and "column" not in list(mapping_required["name"].keys())
    ):
        mapping_required["name"]["column"] = enriched_column_name

    return data_frame, mapping_required


async def instrument_search(
    api_factory: lusid.SyncApiClientFactory, search_requests: list, **kwargs
) -> list:
    """
    Conducts an instrument search for a single instrument

    Parameters
    ----------
    api_factory :       lusid.SyncApiClientFactory
                        The api factory to use
    search_requests:    list[lusid.models.InstrumentSearchProperty]
                        The search requests for this instrument
    kwargs

    Returns
    -------
    None

    """

    instrument_search_results = []

    for search_request in search_requests:
        try:
            result = await instrument_search_single(
                api_factory, search_request, **kwargs
            )
            instrument_search_results.append(result)
        except lusid.exceptions.ApiException as e:
            instrument_search_results.append(e)
    return instrument_search_results


@run_in_executor
def instrument_search_single(
    api_factory: lusid.SyncApiClientFactory,
    search_request: lusid.models.InstrumentSearchProperty,
    **kwargs,
) -> list:
    """
    Conducts an instrument search with a single search request

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        The Api factory to use
    search_request : lusid.models.InstrumentSearchProperty
        The search request
    kwargs

    Returns
    -------
    list[lusid.models.InstrumentMatch]
        The results of the search
    """
    
    api_instance = api_factory.build(lusid.api.SearchApi)
    
    return api_instance.instruments_search(instrument_search_property=[search_request])
                                     



import pandas as pd
from finbourne_sdk_utils.cocoon.cocoon import load_from_data_frame
from finbourne_sdk_utils.cocoon.utilities import load_json_file
import logging
from pathlib import Path

logger = logging.getLogger()

default_mappings = dict(load_json_file("config/seed_sample_data.json"))


def seed_data(
    api_factory,
    domains,
    scope: str,
    transaction_file: str,
    file_type: str,
    mappings: dict = default_mappings,
    sub_holding_keys=[],
):
    """
    This function allows users to seed their LUSID environment with some core data (e.g. instruments,
    portfolios, and transaction) from one file.

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        The api factory to use
    domains : list[str]
        A list of the file_types for upload.
    scope : str
        The scope of the transaction portfolio.
    transaction_file : Union[str, panads.DataFrame]
        The absolute or relative path to source file of transaction data or a pandas DataFrame
    file_type : str
        the file extension (e.g. "csv" for "test_transaction.csv".
    mappings : dict
        a file containing mapping of DataFrame headers to LUSID headers.
    sub_holding_keys : list
        a list of sub-holding keys for grouping.

    Returns
    -------
    overall_results : dict
        An object containing the responses for each domain upload.

    """

    if file_type == "DataFrame" and type(transaction_file) == pd.DataFrame:
        data_frame = transaction_file

    else:

        # Gather a dictionary of supported files

        supported_files = {"csv": "csv", "xlsx": "excel"}

        if Path(transaction_file).suffix != "." + file_type.lower():
            raise ValueError(
                f"""Inconsistent file and file extensions passed: {str(transaction_file)} does not have file extension {file_type}"""
            )

        if file_type not in supported_files:
            raise ValueError(
                f"Unsupported file type, please upload one of the following: {list(supported_files.keys())}"
            )

        data_frame = getattr(pd, f"read_{supported_files[file_type]}")(transaction_file)

    def check_or_set_default_value(mapping, check_key, default_value):
        """
        This function check whether the mappings variables have the required default values to be uploaded
        via the load_from_data_frame function.

        Parameters
        ----------
        mapping : dict
            a file containing mapping of DataFrame headers to LUSID headers.

        check_key : str
            a string that represents the key to be checked

        default_value : obj
            an object such as an empty dictionary or list

        Returns
        -------
        mapping : dict
            a file containing mapping of DataFrame headers to LUSID headers.

        """

        for value in mapping.values():
            if isinstance(value, dict) and check_key not in value.keys():
                value.update({check_key: default_value})

        return mapping

    def generic_load_from_data_frame(file_type):

        return (
            load_from_data_frame(
                api_factory=api_factory,
                file_type=file_type,
                scope=scope,
                data_frame=data_frame,
                mapping_required=mappings[file_type]["required"],
                mapping_optional=mappings[file_type]["optional"],
                identifier_mapping=mappings[file_type]["identifier_mapping"],
                property_columns=mappings[file_type]["properties"],
                properties_scope=scope,
                sub_holding_keys=sub_holding_keys,
            ),
        )

    overall_results = {}

    mappings = check_or_set_default_value(mappings, "optional", {})
    mappings = check_or_set_default_value(mappings, "identifier_mapping", {})
    mappings = check_or_set_default_value(mappings, "properties", [])

    for domain in domains:

        if domain not in mappings:
            raise ValueError(
                f"The provided file_type of {domain} has no associated mapping"
            )

        logging.info(f"Loading {domain} DataFrame into LUSID...")

        response = generic_load_from_data_frame(domain)

        logging.info(f"Loading of {domain} is COMPLETED")

        overall_results[domain] = response

    return overall_results



from setuptools import setup, find_packages  # noqa: H301

version = {}
with open("./__version__.py") as fp:
    exec(fp.read(), version)

setup(
    name="finbourne-sdk-utils",
    version=version["__version__"],
    description="Python utilitiesfor LUSID V2 SDK",
    url="https://github.com/finbourne/finbourne-sdk-utils",
    author="FINBOURNE Technology",
    author_email="engineering@finbourne.com",
    license="MIT",
    keywords=["FINBOURNE", "LUSID", "LUSID SDK", "python"],
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        "urllib3>=1.26.9",
        "requests>=2.27.1",
        "coloredlogs>=14.0",
        "detect_delimiter>=0.1",
        "flatten-json>=0.1.7",
        "pandas>=1.1.4",
        "PyYAML>=5.4",
        "tqdm>=4.52.0",
        "openpyxl>=3.0.7",
        "xlrd~=1.2",
        "pytz>=2019.3",
        "IPython>=7.31.1",
        "lusid-sdk>=2",
    ],
    include_package_data=True,
    python_requires=">=3.7",
)



import finbourne_sdk_utils.cocoon.cocoon
import finbourne_sdk_utils.cocoon.instruments
import finbourne_sdk_utils.cocoon.properties
import finbourne_sdk_utils.cocoon.systemConfiguration
import finbourne_sdk_utils.cocoon.utilities
import finbourne_sdk_utils.cocoon.async_tools
import finbourne_sdk_utils.cocoon.validator
import finbourne_sdk_utils.cocoon.dateorcutlabel
